/*  $Id: more_server.pl, v0.7.0 2005/5/9  zhisheng Exp $


    Author:        Zhisheng Huang
    E-mail:        huang@cs.vu.nl
    WWW:           http://www.cs.vu.nl/~huang
    Copyright (C): 2005 Vrije University Amsterdam

This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    As a special exception, if you link this library with other files,
    compiled with a Free Software compiler, to produce an executable, this
    library does not by itself cause the resulting executable to be covered
    by the GNU General Public License. This exception does not however
    invalidate any other reasons why the executable file might be covered by
    the GNU General Public License.

*/

% this server requires XDIG version 1.4.6 or higher.


:-use_module(library('dig/dig_client')).
:-use_module(library('dig/dig_server')).
:-use_module(library('dig/dig_process')).
:-use_module(library('dig/dig_db')).
:-use_module(library('http/thread_httpd')).

%:-use_module(library('dig/more/more_process')).

:-use_module(library('dig/owl/owl2dig')).



		



%deal with the tellm request
my_dig_server_processing(RequestData, Answer, [connection(close)]):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=tellm,
			!,
                  tellm_processing(Data, Answer),
                  true.



%deal with the askm request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type=askm,
			!,
			askm_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_H, Reply),
                  dig_message_log(more, responseLog, 'ask_response_log.xml', A),
                  true.


%deal with the set request
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = set,
			!,
			dig_setting_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  true.


%deal with the do operation
my_dig_server_processing(RequestData, Answer, _Options):-
			dig_requestdata_analysis(RequestData, Data, Type),
			Type = do,
			!,
			dig_do_processing(Data, A),
                  Reply = [element(response, [], A)],
                  Answer = answer(_, Reply),
                  dig_message_log(more, responseLog, 'do_response_log.xml', A),
                  true.




%deal with other request
 my_dig_server_processing(RequestData, Answer, Options):-
			dig_post(RequestData, Answer, Options).



% tellm_processing(+Data, -Answer)
% where Answer = answer(Header, ElementList)

% have dealed with the whole list, returning the OK message.
tellm_processing(Data, Answer):-
 %  format('Content-type: text/plain~n~n',[]),
    dig_get_element(Data, E, _Other),
    E = nil,
    !,
    Answer = answer(_H, text('<response><ok>tellm</ok></response>')).


% deal with an ontology statement which refers to a URL
tellm_processing(Data, Answer):-
    dig_get_element(Data, E, Other),
    E = element(ontology, Attr, _Ontodata),   
    dig_select_attribute(Attr, url, URL),
    URL \=nil,
    !,
    dig_select_attribute(Attr, name, ID),
    dig_select_attribute(Attr, language, Lang),
    more_load_ontology(ID, Lang, dig, URL),   
    tellm_processing(Other, Answer).


% deal with an ontology statement, then recursively  process other
tellm_processing(Data, Answer):-
    dig_get_element(Data, E, Other),
    E = element(ontology, Attr, Ontodata),
    !,
    dig_select_attribute(Attr, name, ID),
    dig_select_attribute(Attr, language, Lang),
    more_add_ontology(ID, Lang, dig, Ontodata),
    tellm_processing(Other, Answer).

% deal with an versionSpace statement, then recursively  process other
tellm_processing(Data, Answer):-
    dig_get_element(Data, E, Other),
    E = element(versionSpace, [name=C], CData),
    !,
    versionSpace_processing(C, CData),
    tellm_processing(Other, Answer).



% askm_processing(+Data, -Reply)
% where Reply = an element, like element(true, [id=ID], [])


% deal with the case C |= q

askm_processing(Data, Reply):-
    dig_get_element(Data, E, Other),
    E = element(querym, Attr, Qdata),    
    dig_select_attribute(Attr, ontology, O),    
    dig_select_attribute(Attr, id, ID),
    dig_select_attribute(Attr, versionSpace, C),
    O  = nil,
    !,
    versionSpace_get_ontology(C, latest, O1),
    more_query_evaluation(C, O1, ID, Qdata, Reply1),
    askm_processing(Other, Replies),
    Reply = [Reply1 | Replies],
    true.

% deal with the case C, o |= q

askm_processing(Data, Reply):-
    dig_get_element(Data, E, Other),
    E = element(querym, Attr, Qdata),    
    dig_select_attribute(Attr, ontology, O),    
    dig_select_attribute(Attr, id, ID),
    dig_select_attribute(Attr, versionSpace, C),
    !,
    more_query_evaluation(C, O, ID, Qdata, Reply1),
    askm_processing(Other, Replies),
    Reply = [Reply1 | Replies],
    true.


   
askm_processing(_Data, Reply):-
    !,
    Reply = [].


more_load_ontology(ID, Lang, dig, URL):-
   (Lang = dig; Lang = owl),
   !,
   KB = kb(more, ID),
   more_clear_tmp(ID),
   dig_load_url_telldata(URL, KB).

   
more_load_ontology(ID,  Lang, _, _URL):-
    format(user_error, 'unknown ontology language ~w for ~w~n', [Lang, ID]). 


%more_add_ontology(+OntologyName, +CurrentDataFormat, +InternalDataFormat, +OntologySpecification)

% dig as the internal data format
% no convertion
more_add_ontology(ID, dig, dig, OntoData):-
   more_clear_tmp(ID),
   dig_assert_data(kb(more,ID), OntoData).


% converting owl into dig
more_add_ontology(ID, owl, dig, OntoData):-
   owl2dig(elements(OntoData), elements(L)),
   E = element(tells, _, L1),
   dig_get_element(L, E, _),
   more_clear_tmp(ID),
   dig_assert_data(kb(more,ID), L1). 

  
more_add_ontology(ID,  Lang, _, _OntoData):-
    format(user_error, 'unknown ontology language ~w for ~w~n', [Lang, ID]). 


% clear all tmp information
   more_clear_tmp(ID):-
	retractall(dig_db_element(tmp(more, ID), _)),
      !.

  more_clear_tmp(_ID):-!.
        

% deal with versionSpace statements

% versionSpace statments have been done
versionSpace_processing(_C, CData):-
  dig_get_element(CData, E, _Other),
  E = nil,
  !.
 
% deal with the pair statement 
versionSpace_processing(C, CData):-
  dig_get_element(CData, E, Other),
  E = element(pair, Attr, []),
  !,
  dig_select_attribute(Attr, ontology1, O1),
  dig_select_attribute(Attr, ontology2, O2),
  assert(dig_db_element(setting(more, versionSpace(C)), pair(O1,O2))),
  versionSpace_processing(C, Other).

% deal with the nopair statement
versionSpace_processing(C, CData):-
  dig_get_element(CData, E, Other),
  E = element(nopair, Attr, []),
  !,
  dig_select_attribute(Attr, ontology1, O1),
  dig_select_attribute(Attr, ontology2, O2),
  retract(dig_db_element(setting(more, versionSpace(C)), pair(O1,O2))),
  versionSpace_processing(C, Other).


% ignoring other cases
versionSpace_processing(_,_):-!.


  
%more_query_evaluation(+versionSpace, +Ontology, +QueryID, +QueryData, -Reply)
%preprocessing a query

%call the query evaluation without the query ID
more_query_evaluation(C, O, ID, QData, Reply):-
      more_query_evaluation(C, O, QData, Status),
      !,
      more_get_reply(Status, ID, Reply),
    	true.


% display the unknown query format
more_query_evaluation(C, _O, ID, QData, Reply):-
	Reply = element(unKnownQuery, [versionSpace=C, id=ID], QData),
      true.

%more_get_reply(+Status, +ID, -Reply)
%get the correct reply format

more_get_reply(Status, ID, Reply):-
    Status = status(Feature, L),
    !,
    Reply = element(Feature, [id=ID], L).


more_get_reply(Status, ID, Reply):-
	Reply = element(Status, [id=ID], []).



%more_query_evaluation(+versionSpace, +Ontology, +QueryData, -Reply)
%deal with a single query


% deal with the primitive query
more_query_evaluation(_C, O, QData, Status):-
	select(element(query, _Attr,  Q),QData, _),
    	!,
       dig_db_post(kb(more,O), withClearKB, _Answer1, [connection(close)]),
    	dig_ask(elements(Q), Answer, [connection(close)]),
  	Answer = answer(_H, Response),
	dig_response_analysis(Response, Status1, _Element, _Other),
      status_evaluation(Status1, Status),
      true.


% conjunction
more_query_evaluation(C, O, QData, Status):-
	select(element(and, _Attr, Q),QData, _),
      dig_get_element(Q, E1, Q1),
      !,
      more_query_evaluation(C, O, [E1], Status1),
      more_query_evaluation(C, O, Q1, Status2),
      status_evaluation(and, Status1, Status2, Status). 


% disjunction
more_query_evaluation(C, O, QData, Status):-
	select(element(or, _Attr, Q),QData, _),
      dig_get_element(Q, E1, Q1),
      !,
      more_query_evaluation(C, O, [E1], Status1),
      more_query_evaluation(C, O, Q1, Status2),
      status_evaluation(or, Status1, Status2, Status).

% implication
more_query_evaluation(C, O, QData, Status):-
	select(element(implies, _Attr, Q),QData, _),
      dig_get_element(Q, E1, Q1),
      !,
      more_query_evaluation(C, O, [E1], Status1),
      more_query_evaluation(C, O, Q1, Status2),
      status_evaluation(implies, Status1, Status2, Status).

% biconditional
more_query_evaluation(C, O, QData, Status):-
	select(element(iff, _Attr, Q),QData, _),
      dig_get_element(Q, E1, Q1),
      !,
      more_query_evaluation(C, O, [E1], Status1),
      more_query_evaluation(C, O, Q1, Status2),
      status_evaluation(iff, Status1, Status2, Status).

% negation
more_query_evaluation(C, O, QData, Status):-
	select(element(not, _Attr, Q), QData, _),
      !,
      more_query_evaluation(C, O, Q, Status1),
      status_evaluation(not, Status1, Status). 

% the previous operator

more_query_evaluation(C, O, QData, Status):-
      (Term = previousVersion;Term = previous),
	select(element(Term, _Attr, Q), QData, _),
      !,
      more_temporalquery_evaluation(previous, C, O, Q, Status).

% the alwaysp operator
more_query_evaluation(C, O, QData, Status):-
       (Term =allPriorVersions; Term = alwaysinpast; Term = alwaysp),
	  select(element(Term, _Attr, Q), QData, _),
		      !,
      more_temporalquery_evaluation(alwaysp, C, O, Q, Status).

% the past operator 
more_query_evaluation(C, O, QData, Status):-
      (Term = somePriorVersion; Term = past),
	select(element(Term, _Attr, Q), QData, _),
      !,
      more_temporalquery_evaluation(past, C, O, Q, Status).

% the allversion operator 
more_query_evaluation(C, O, QData, Status):-
      Term = allVersions,
	select(element(Term, _Attr, Q), QData, _),
      !,
      more_temporalquery_evaluation(allversion, C, O, Q, Status).

% the since operator 
more_query_evaluation(C, O, QData, Status):-
      Term = since,
	select(element(Term, _Attr1, Q), QData, _),
      dig_get_element(Q, Q1, Q0),
      dig_get_element(Q0, Q2, _),
      !,
      more_temporalquery_evaluation(since0, C, O, [[Q1], [Q2]], Status).


% the new/obsolete/invariant related concept query

more_query_evaluation(C, O, QData, Status):-
      (S = newOnConcept; S = obsoleteOnConcept; S=invariantOnConcept),
	select(element(S, Attrib, L1),QData, _),
      L1=[],
      dig_select_attribute(Attrib, concept, Concept),
      Concept = nil,
      !,
      KBID= kb(more, O),
      dig_all_conceptname(KBID, AllConceptNames1),
      allconceptname(AllConceptNames1, AllConceptNames),
      dig_select_attribute(Attrib, type, Type),
      dig_select_attribute(Attrib, comparedOntology, O1),
      more_get_concepts_list(S, C, O, AllConceptNames, Type, O1, L),
      Status= status(answer, L), 
      true.


more_query_evaluation(C, O, QData, Status):-
      (S = newOnConcept; S = obsoleteOnConcept; S=invariantOnConcept),
	select(element(S, Attrib, L1),QData, _),
      L1=[],
    	!,
      dig_select_attribute(Attrib, concept, Concept),
      dig_select_attribute(Attrib, type, Type),
      dig_select_attribute(Attrib, comparedOntology, O1),
      more_get_concepts(S, C, O, Concept, Type, O1, L),
      Status= status(answer, L), 
      true.



more_query_evaluation(C, O, QData, Status):-
      (S = newOnConcept; S = obsoleteOnConcept; S=invariantOnConcept),
	select(element(S, Attrib, L1), QData, _),
    	!,
      dig_select_attribute(Attrib, type, Type),
      dig_select_attribute(Attrib, comparedOntology, O1),
      more_get_concepts_list(S, C, O, L1, Type, O1, L),
      Status= status(answer, L),   
      true.

% version retrivial

more_query_evaluation(C, O, QData, Status):-
	select(element(ontologies, _Attr,  Q),QData, _),
    	!,
      more_version_retrivial(C, O, Q, L),
      Status = status(ontologySet, L),
      true.

% version numbering
more_query_evaluation(C, O, QData, Status):-
      S = version,
	select(element(S, Attrib, L), QData, _),
    	!,
      dig_select_attribute(Attrib, number, N1),
      atom_number(N1, N),
      more_version_numbering(C, O, N, L, Status),  
      true.

% version numbering processing
more_version_numbering(C, O, N, L, Status):-
   N = 0,
   !,
   more_query_evaluation(C, O, L, Status).

% valid relative version numbering
more_version_numbering(C, O, N, L, Status):-
   N < 0,
   dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
   !,
   N1 is 1 + N,
   more_version_numbering(C, O1, N1, L, Status).

% invalid relative version numbering
more_version_numbering(_C, _O, N, _L, Status):-
   N < 0,
   !,
   Status = false.

% absolute version numbering

more_version_numbering(C, _O, N, L, Status):-
  !, 
  versionSpace_get_information(C, _O1, On, _Os, No),
  N1 is N - No,
  more_version_numbering(C, On, N1, L, Status).
  
  



 



% version retrivial processing

more_version_retrivial(C, O, Q, L):-   
   more_query_evaluation(C, O, Q, Status), 
   dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
   !,
   more_version_status(O, Status, E), 
   more_version_retrivial(C, O1, Q, L1),
   L = [E|L1].

more_version_retrivial(_C, _O, _Q, L):-
    !,
    L =[].

more_version_status(O, Status, E):-
  Status = true,
  !,
  E = element(ontology, [name=O], []).

more_version_status(_O, _Status, E):-
  !,
  E = ''.  

%more_get_concepts_list(+S, +versionSpace, +Ontology, +ConceptList, +Type, +ComparedOntology, -ConceptSets)
%S = newOnConcept; S = obsoleteOnConcept; S=invariantOnConcept,
%Type = children, parents, etc.

more_get_concepts_list(_S, _C, _O, [], _Type, _O1, _L):-
    !.


more_get_concepts_list(S, C, O, [X|Y], Type, O1, L):-
     X = element(catom, [name=Concept], []), 
     !,
    more_get_concepts_list(S, C, O, Y, Type, O1, L21),
    more_get_concepts(S, C, O, Concept, Type, O1, L22),
    (L22=['NIL'] -> X1 ='';X1 = element(conceptRelated, [concept=Concept], L22)),
    concat_atom([log, '/', S, '_', O, '_', Concept, '_', Type, '_', O1, '.xml'], File0),
    dig_message_log(more, conceptLog, File0, [X1]),
    L = [X1|L21].

more_get_concepts_list(S, C, O, [_X|Y], Type, O1, L):-
     !,
    more_get_concepts_list(S, C, O, Y, Type, O1, L).


%more_get_concepts(+S, +versionSpace, +Ontology, +Concept, +Type, +ComparedOntology, -ConceptSets)

%If it does not state any type, then deal with all of the types.
more_get_concepts(S, C, O1, Concept, Type, O2, L):-
     Type = nil,
     !,
     more_get_concepts_type(S, C, O1, Concept, children, O2, E1),
     more_get_concepts_type(S, C, O1, Concept, parents, O2, E2),
     more_get_concepts_type(S, C, O1, Concept, ancestors, O2, E3),
     more_get_concepts_type(S, C, O1, Concept, descendants, O2, E4),
     more_get_concepts_type(S, C, O1, Concept, equivalents, O2, E5),
     L1 = [E1, E2, E3, E4, E5],
     make_emptylist_nil(L1, L), 
     true.

% Type is not nil
more_get_concepts(S, C, O1, Concept, Type, O2, L):-
	more_get_concepts_type(S, C, O1, Concept, Type, O2, E),
%     L = [E],
      make_emptyelement_nil(E, L),
      true.

 make_emptylist_nil(L1, L):- 
	dig_get_element(L1, E, _),
      E = nil,
      !,
      L = ['NIL'].

make_emptylist_nil(L1, L):- 
    L1 =L.


 make_emptyelement_nil(E, L):-
    E = '',
    !,
    L = ['NIL'].

 make_emptyelement_nil(E, L):-
    L = [E].


% more_get_concepts_type(+S, +versionSpace, +Ontology, +Concept, +Type, +ComparedOntology, -ConceptSetElement)

%if the compared ontology is nil, then get the previous one
more_get_concepts_type(newOnConcept, C, O1, Concept, Type, O2, E):-
     O2 = nil,
     dig_db_element(setting(more, versionSpace(C)), pair(O, O1)),
     !,
     ID1= kb(more, O1),
     ID2= kb(more, O),
     more_related_conceptname(ID1, Concept, Type, L1),
     more_related_conceptname(ID2, Concept, Type, L2),   
     subtract(L1, L2, L),
%     E = element(Type, [], L),
     make_emptylist_nil(L, Type, E),
     true.

%if the previous one does not exists, then it is initial version at the versionSpace 
more_get_concepts_type(newOnConcept, _C, O1, Concept, Type, O2, E):-
     O2 = nil,
     !,
     ID1= kb(more, O1),
     more_related_conceptname(ID1, Concept, Type, L1),
%     E = element(Type, [], L1),
     make_emptylist_nil(L1, Type, E),
     true.


more_get_concepts_type(newOnConcept,_C, O1, Concept, Type, O2, E):-
     ID1= kb(more, O1),
     ID2= kb(more, O2),
     more_related_conceptname(ID1, Concept, Type, L1),
     more_related_conceptname(ID2, Concept, Type, L2),
     !,   
     subtract(L1, L2, L),
%     E = element(Type, [], L),
     make_emptylist_nil(L, Type, E),
     true.


more_get_concepts_type(obsoleteOnConcept, C, O1, Concept, Type, O2, E):-
     O2 = nil,
     dig_db_element(setting(more, versionSpace(C)), pair(O, O1)),
     !,
     more_get_concepts_type(obsoleteOnConcept, C, O1, Concept, Type, O, E).

more_get_concepts_type(obsoleteOnConcept, _C, _O1, _Concept, _Type, O2, E):-
     O2 = nil,
     !,
 %    E = element(Type, [], []),
     E = '',
     true.


more_get_concepts_type(obsoleteOnConcept, C, O1, Concept, Type, O2, E):-
     more_get_concepts_type(newOnConcept, C, O2, Concept, Type, O1, E).


more_get_concepts_type(invariantOnConcept, C, O1, Concept, Type, O2, E):-
     O2 = nil,
     dig_db_element(setting(more, versionSpace(C)), pair(O, O1)),
     !,
     more_get_concepts_type(invariantOnConcept, C, O1, Concept, Type, O, E).

more_get_concepts_type(invariantOnConcept, _C, _O1, _Concept, _Type, O2, E):-
     O2 = nil,
     !,
 %    E = element(Type, [], []),
     E = '',
     true.


more_get_concepts_type(invariantOnConcept,_C, O1, Concept, Type, O2, E):-
     ID1= kb(more, O1),
     ID2= kb(more, O2),
     more_related_conceptname(ID1, Concept, Type, L1),
     more_related_conceptname(ID2, Concept, Type, L2),
     !,   
     intersection(L1, L2, L),
%    E = element(Type, [], L),
     make_emptylist_nil(L, Type, E),
     true.

% check if the data is available in the temp database
more_related_conceptname(ID1, Concept, Type, L1):-
     ID1 = kb(more, O1),
     dig_db_element(tmp(more, O1), [Concept, Type, L]),
     !,
     L1 = L.
     
% otherwise get the data and save it in the temp database
more_related_conceptname(ID1, Concept, Type, L1):-
 !,
 ID1 = kb(more, O1),
 dig_related_conceptname(ID1, Concept, Type, L1),
 assert(dig_db_element(tmp(more, O1), [Concept, Type, L1])).



make_emptylist_nil(L, _, E):-
   L = [],
   !,
   E = ''.

make_emptylist_nil(L, Type, E):-
    E = element(Type, [], L).




% deal with temporal formulas evaluation

more_temporalquery_evaluation(previous, C, O, Q, Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     !,
     more_query_evaluation(C, O1, Q, Status).

more_temporalquery_evaluation(previous, _C, _O, _Q, Status):-
     !,
     Status =false.


more_temporalquery_evaluation(alwaysp, C, O, Q, Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     more_query_evaluation(C, O1, Q, Status1),
     (Status1 = false; Status1 =error),
     !,
     Status = false.



more_temporalquery_evaluation(alwaysp, C, O, Q, Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     !,
     more_temporalquery_evaluation(alwaysp, C, O1, Q, Status).


more_temporalquery_evaluation(alwaysp, _C, _O, _Q, Status):-
      !,
	Status = true.


more_temporalquery_evaluation(past, C, O, Q, Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     more_query_evaluation(C, O1, Q, Status1),
     Status1 = true,
     !,
     Status = Status1.



more_temporalquery_evaluation(past, C, O, Q, Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     !,
     more_temporalquery_evaluation(past, C, O1, Q, Status).


more_temporalquery_evaluation(past, _C, _O, _Q, Status):-
      !,
	Status = false.


more_temporalquery_evaluation(allversion, C, O, Q, Status):-
     more_query_evaluation(C, O, Q, Status1),
     (Status1 = false; Status1 =error),
     !,
     Status = false.



more_temporalquery_evaluation(allversion, C, O, Q, Status):-
     !,
     more_temporalquery_evaluation(alwaysp, C, O, Q, Status).

% evaluation of the since operator

% initial processing of the since operator 
more_temporalquery_evaluation(since0, C, O, [Q1, Q2], Status):-
     more_query_evaluation(C, O, Q1, Status1),
     more_query_evaluation(C, O, Q2, Status2),
     !,
     more_since_evaluation(C, O, [Q1, Q2], Status1, Status2, Status).



% normal processing of the since operator
more_temporalquery_evaluation(since, C, O, [Q1, Q2], Status):-
     dig_db_element(setting(more, versionSpace(C)), pair(O1, O)),
     more_query_evaluation(C, O1, Q1, Status1),
     more_query_evaluation(C, O1, Q2, Status2),
     !,
     more_since_evaluation(C, O1, [Q1, Q2], Status1, Status2, Status).


more_temporalquery_evaluation(since, _C, _O, _Q, Status):-
      !,
	Status = false.


more_since_evaluation(_C, _O, _Q, Status1, _Status2, Status):-
      (Status1 = false; Status1 =error),
       !,
       Status = false.

more_since_evaluation(_C, _O, _Q, _Status1, Status2, Status):-
      Status2 = true, 
       !,
       Status = true.

more_since_evaluation(C, O, Q, _Status1, _Status2, Status):-
      !,
	more_temporalquery_evaluation(since, C, O, Q, Status).



% versionSpace_get_information(+versionSpace, -initialOntology, -latestOntology, -OntologyList, -OntologyNumber)


versionSpace_get_information(C, Oi, Ol, L, N):-
  dig_db_element(setting(more, versionSpace(C)), pair(O1, O2)),
  !,
  versionSpace_get_ontology(C, initial, O1, Oi, L1),
  versionSpace_get_ontology(C, latest, O2, Ol, L2),
  reverse(L1, L3),
  append(L3, L2, L),
  length(L, N).

versionSpace_get_information(_C, Oi, Ol, L, N):-
  !,
  Oi=nil,
  Ol=nil,
  L = [],
  N = 0.
  
  
  




% versionSpace_get_ontology(+versionSpace, +Type, -OntologyName)

versionSpace_get_ontology(C, Type, O):-
      (dig_db_element(setting(more, versionSpace(C)), pair(O1, _O2));  dig_db_element(setting(more, versionSpace(C)), pair(_O2, _O1))),
      !,
	versionSpace_get_ontology(C, Type, O1, O, _L).

versionSpace_get_ontology(_C, _Type, O):-
     !,
     O = nil.



% recursively search for the latest in a linear versionSpace.
versionSpace_get_ontology(C, latest, O1, O, L):-
    dig_db_element(setting(more, versionSpace(C)), pair(O1, O2)),
    !,
    versionSpace_get_ontology(C, latest, O2, O, L2),
    (member(O1, L2)->L = L2; L =[O1|L2]).

% O1 is the latest one in the versionSpace.
versionSpace_get_ontology(C, latest, O1, O, L):-
    dig_db_element(setting(more, versionSpace(C)), pair(_O2, O1)),
    !,
    O = O1,
    L = [O1].

% recursively search for the initial in a linear versionSpace.
versionSpace_get_ontology(C, initial, O1, O, L):-
    dig_db_element(setting(more, versionSpace(C)), pair(O2, O1)),
    !,
    versionSpace_get_ontology(C, initial, O2, O, L2),
    (member(O1, L2)->L = L2; L =[O1|L2]).

% O1 is the initial one in the versionSpace.
versionSpace_get_ontology(C, initial, O1, O, L):-
    dig_db_element(setting(more, versionSpace(C)), pair(O1, _O2)),
    !,
    O = O1,
    L = [O1].


% empty versionSpace
versionSpace_get_ontology(C, _Type, _O1, O, L):-
   format(user_error, 'empty versionSpace ~w specification~n', [C]),
   O = nil,
   L = [].

   
    
   


/* old evaluation format, keep the error from the false
status_evaluation(and, true, true, true):-!.
status_evaluation(and, error, _, error):-!.
status_evaluation(and, _, error, error):-!.
status_evaluation(and, false, _, false):-!.
status_evaluation(and, _, false, false):-!.

status_evaluation(or, false, false, false):-!.
status_evaluation(or, true, _, true):-!.
status_evaluation(or, _, true, true):-!.
status_evaluation(or, error, _, error):-!.
status_evaluation(or, _, error, error):-!.

status_evaluation(iff, false, false, true):-!.
status_evaluation(iff, true, true, true):-!.
status_evaluation(iff, error, _, error):-!.
status_evaluation(iff, _, error, error):-!.
status_evaluation(iff, _, _, false):-!.

status_evaluation(implies, true, false, false):-!.
status_evaluation(implies, error, _, error):-!.
status_evaluation(implies, _, error, error):-!.
status_evaluation(implies, _, _, true):-!.

status_evaluation(_, _, _, error):-
 format('unknown status evaluation~n', []).


status_evaluation(not, true, false):-!.
status_evaluation(not, false, true):-!.
status_evaluation(not, _, error):-!.
*/

% convert the error into the false

status_evaluation(Status1, Status):-
    Status1 = error,
    !,
    Status = false.

% otherwise unchange it
status_evaluation(Status1, Status):-
   Status1 = Status.




status_evaluation(and, true, true, true):-!.
status_evaluation(and, false, _, false):-!.
status_evaluation(and, _, false, false):-!.

status_evaluation(or, false, false, false):-!.
status_evaluation(or, true, _, true):-!.
status_evaluation(or, _, true, true):-!.

status_evaluation(iff, false, false, true):-!.
status_evaluation(iff, true, true, true):-!.
status_evaluation(iff, _, _, false):-!.

status_evaluation(implies, true, false, false):-!.
status_evaluation(implies, _, _, true):-!.

status_evaluation(_, _, _, error):-
 format('unknown status evaluation~n', []).


status_evaluation(not, true, false):-!.
status_evaluation(not, false, true):-!.


status_evaluation(_, _, error):-
 format('unknown status evaluation~n', []).

  
more_get_server_port(Port):-
  dig_db_element(setting(more, serverPort), Port).

more_set_server_port(Port):-
  retractall(dig_db_element(setting(more, serverPort), _)), 
  assert(dig_db_element(setting(more, serverPort), Port)).  

more_default_server_port(8003):-!.

more_server_init(Port):-
   more_default_server_port(Port),
   more_set_server_port(Port),
   true.

more_setting(X, Y):-
   dig_db_element(setting(more, X), Y).

more_init:-
   retractall(dig_db_element(setting(more, _), _)), 
   assert(dig_db_element(setting(more, conceptLog), off)),
   assert(dig_db_element(setting(more, responseLog), on)),
%  assert(dig_db_element(setting(more, versionSpace), [])),
   more_server_init(_Port),
   true.
   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  message log and debug
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%dig_message_log(+SystemID, +LogType, +FileName, +Elements),

dig_message_log(S, LogType, File0, L):-
    dig_db_element(setting(S, LogType), on),
    L \=[''],
    !,
    open(File0, write, Out, []),
    dig_xmlns(XMLNS),
    TellElements = [element(response, XMLNS, L)],
    elements_xmltext(TellElements, file(File0)),
    close(Out),
    !.

        
dig_message_log(_S, _LogType, _F, _L):-!.



   
			

:- more_init, more_get_server_port(Port), set_dig_server_id('SEKT MORE Server 0.5.0'), http_server(dig_server,[port(Port)]).




	   
     